#ifndef DUNE_REDUCEDROD_CELLASSEMBLER_HH
#define DUNE_REDUCEDROD_CELLASSEMBLER_HH

#include <dune/common/parametertree.hh>

#include <dune/istl/matrixindexset.hh>

#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>

using namespace Dune;
using namespace Functions;


template <class Basis> //, class LocalScalar, class Local2Tensor> // LocalFunction derived from basis?
class CellAssembler {

public:
	static const int nCompo = 3; //GridView::dimensionworld;
	static const int dim = 3; // typename Basis::GridView::dimension
	// enum {dim=Basis::LocalView::GridView::dimension};

	using GridView = typename Basis::GridView;

	using Domain = typename GridView::template Codim<0>::Geometry::GlobalCoordinate;

	using ScalarRT = FieldVector< double, 1>;
	using VectorRT = FieldVector< double, nCompo>;
	using MatrixRT = FieldMatrix< double, nCompo, nCompo>;

	using FuncScalar = std::function< ScalarRT(const Domain&) >;
	using Func2Tensor = std::function< MatrixRT(const Domain&) >;

	using VectorCT = BlockVector<FieldVector<double,1> >;
	using MatrixCT = BCRSMatrix<FieldMatrix<double,1,1> >;
	using ElementMatrixCT = Matrix<FieldMatrix<double,1,1> >;


protected:
//private:

	const Basis& basis_; // Basis kopieren?

	std::fstream& log_;
	const ParameterTree& parameterSet_;

	const FuncScalar mu_; //ref? const?
    const FuncScalar lambda_;
    double gamma_;

    std::map<int,int> perIndexMap_;
    int sizeBasisTorus_;

    // kann fest im objekt enthalten sein
    /*
    Grid2Tensor E_a_;
    Grid2Tensor E_K1_;
    Grid2Tensor E_K2_;
    Grid2Tensor E_K3_; */


public:

	// Constructor
	CellAssembler(  const Basis& basis,
					const FuncScalar& mu,
					const FuncScalar& lambda,
					double gamma,
					std::fstream& log,
					const ParameterTree& parameterSet)
        : basis_(basis),
          mu_(mu),
          lambda_(lambda),
          gamma_(gamma),
          log_(log),
          parameterSet_(parameterSet)
    {
    	//periodic FE index mapper
    	sizeBasisTorus_ = assemblePeriodicIndexMap();
    }

	// getter

	//Basis getBasis() const {return basis_;}
	const Basis* getBasis() {return &basis_;}

	std::map<int,int> getIndexMapPeriodicBoundary(){return perIndexMap_;}

    ParameterTree getParameterSet() const {return parameterSet_;}

    std::fstream* getLog(){return &log_;}

    int getSizeTorusElements(){return sizeBasisTorus_;}

	// setter

	template <class LocalScalar1, class LocalScalar2>
	void computeLocalStiffnessMatrix(const typename Basis::LocalView& localView, const LocalScalar1& mu, const LocalScalar2& lambda, ElementMatrixCT& elementMatrix)
	{// Get the grid element from the local FE basis view
	  auto element = localView.element();
	  auto geometry = element.geometry();

	  const auto& localFiniteElement = localView.tree().child(0).finiteElement();
	  const auto nSf = localFiniteElement.localBasis().size();

	  elementMatrix.setSize(localView.size(), localView.size());
	  elementMatrix = 0;
	  int order = 2*(dim*localFiniteElement.localBasis().order()-1); // für simplex
	  const QuadratureRule< double, dim>& quad = QuadratureRules<double, dim>::rule(element.type(), order);

	  for (const auto& quadPoint : quad){
	  	const Domain& quadPos = quadPoint.position();
	    const auto jacobian = geometry.jacobianInverseTransposed(quadPos);
	    const double integrationElement = geometry.integrationElement(quadPos);

	    std::vector< FieldMatrix< double, 1, nCompo> > referenceGradients;
	    localFiniteElement.localBasis().evaluateJacobian(quadPos, referenceGradients);

	    std::vector< VectorRT> gradients(referenceGradients.size());
	    for (size_t i=0; i< gradients.size(); i++)
	      jacobian.mv(referenceGradients[i][0], gradients[i]);

	    for (size_t i=0; i < nSf; i++)
	      for (size_t j=0; j < nSf; j++ )
	        for (size_t k=0; k < nCompo; k++)
	          for (size_t l=0; l< nCompo; l++)
	          {
	            // Deformation gradient of the ansatz basis function, Domain and Range components are swaped
	            MatrixRT defGradientU(0);
	            defGradientU[k][0] = gradients[i][0];              // Y
	            defGradientU[k][1] = 1.0/gamma_*gradients[i][1];   //X2
	            defGradientU[k][2] = 1.0/gamma_*gradients[i][2];   //X3

	            // Deformation gradient of the test basis function, Domain and Range components are swaped
	            MatrixRT defGradientV(0);
	            defGradientV[l][0] = gradients[j][0];              // Y
	            defGradientV[l][1] = 1.0/gamma_*gradients[j][1];   //X2
	            defGradientV[l][2] = 1.0/gamma_*gradients[j][2];   //X3

	            // Elastic Strains
	            MatrixRT strainU, strainV;
	            for (int ii=0; ii<nCompo; ii++)
	              for (int jj=0; jj<nCompo; jj++)
	              {
	                strainU[ii][jj] = 0.5 * (defGradientU[ii][jj] + defGradientU[jj][ii]);         // symmetric gradient?!
	                strainV[ii][jj] = 0.5 * (defGradientV[ii][jj] + defGradientV[jj][ii]);         // same ?
	              }

	            // St.Venant-Kirchhoff stress
	            MatrixRT stressU(0);
	            stressU.axpy(2*mu(quadPos), strainU); //this += 2mu *strainU

	            double trace = 0;
	            for (int ii=0; ii<nCompo; ii++)
	              trace += strainU[ii][ii];

	            for (int ii=0; ii<nCompo; ii++)
	              stressU[ii][ii] += lambda(quadPos) * trace;

	              // Local energy density: stress times strain
	            double energyDensity = 0;
	            for (int ii=0; ii<nCompo; ii++)
	              for (int jj=0; jj<nCompo; jj++)
	                energyDensity += stressU[ii][jj] * strainV[ii][jj];

	            size_t row = localView.tree().child(k).localIndex(i);
	            size_t col = localView.tree().child(l).localIndex(j);
	            elementMatrix[row][col] += energyDensity * quadPoint.weight() * integrationElement;
	          }
  	}
  }

template<class LocalScalar1, class LocalScalar2, class Local2Tensor>
void computeLocalLoad(const typename Basis::LocalView& localView, const LocalScalar1& mu, const LocalScalar2& lambda, const Local2Tensor& strainE, VectorCT& elementRhs)
{// Get the grid element from the local FE basis view

	elementRhs.resize(localView.size());
  	elementRhs = 0;
  	const auto element = localView.element();
  	const auto geometry = element.geometry();

  	const auto& localFiniteElement = localView.tree().child(0).finiteElement();
  	const auto nSf = localFiniteElement.localBasis().size();

  	int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1); // für simplex
  	//log << "orderQR: " << orderQR << std::endl;
  	const QuadratureRule<double, dim>& quad = QuadratureRules<double, dim>::rule(element.type(), orderQR);

  	for (const auto& quadPoint : quad){
  		const Domain& quadPos = quadPoint.position();
    	const auto jacobian = geometry.jacobianInverseTransposed(quadPos);
    	const double integrationElement = geometry.integrationElement(quadPos);

    	std::vector<FieldMatrix<double,1,nCompo> > referenceGradients;
    	localFiniteElement.localBasis().evaluateJacobian(quadPos, referenceGradients);

    	std::vector< VectorRT > gradients(referenceGradients.size());
    	for (size_t i=0; i< gradients.size(); i++)
      		jacobian.mv(referenceGradients[i][0], gradients[i]);

    	for (size_t i=0; i < nSf; i++)
      		for (size_t k=0; k < nCompo; k++)
      		{
        		// Deformation gradient of the ansatz basis function, 			no: Domain and Range components are swaped
        		MatrixRT defGradientU(0);
        		defGradientU[k][0] = gradients[i][0];              // Y
        		defGradientU[k][1] = 1.0/gamma_*gradients[i][1];   //X2
        		defGradientU[k][2] = 1.0/gamma_*gradients[i][2];   //X3

        		//log_ << defGradientU << std::endl;
        		// Elastic Strain
        		MatrixRT strainU, strainV;                                       //strainV never used?
        		for (int ii=0; ii<nCompo; ii++)
          			for (int jj=0; jj<nCompo; jj++)
            			strainU[ii][jj] = 0.5 * (defGradientU[ii][jj] + defGradientU[jj][ii]);

        		// St.Venant-Kirchhoff stress
        		MatrixRT stressU(0);
        		stressU.axpy(2*mu(quadPos), strainU); //this += 2mu *strainU

        		double trace = 0;
        		for (int ii=0; ii<nCompo; ii++)
          			trace += strainU[ii][ii];

        		for (int ii=0; ii<nCompo; ii++)
          			stressU[ii][ii] += lambda(quadPos) * trace;
          		//log_ << "stressU:\n" << stressU << std::endl;
          		// Local energy density: stress times strain
        		double energyDensity = 0;
        		for (int ii=0; ii<nCompo; ii++)
          			for (int jj=0; jj<nCompo; jj++)
            			energyDensity += stressU[ii][jj] * strainE(quadPos)[ii][jj];
            	//log_ << "strainE:\n" << strainE(quadPoint.position()) << std::endl;

        		size_t row = localView.tree().child(k).localIndex(i);
        		elementRhs[row] += energyDensity * quadPoint.weight() * integrationElement;
        		//log_ << "\nk: " << k << "\ni: " << i << "\nrow: " << row << std::endl;

        //log << "energyDensity quadPoint.weight() integrationElement: " << energyDensity << " " << quadPoint.weight() << " " << integrationElement << std::endl;
      } //end for k
  } // end for quadPoint

}// end method



	void assembleStiffnessMatrix(MatrixCT& matrix)
	{
	    std::cout << "assemble stiffnessmatrix begins.\n";
	    matrix = 0;
	    auto localView = basis_.localView();

	    auto muGridF  = makeGridViewFunction(mu_, basis_.gridView());
	    auto muLocal = localFunction(muGridF);
	    auto lambdaGridF  = makeGridViewFunction(lambda_, basis_.gridView());
	    auto lambdaLocal = localFunction(lambdaGridF);

	  	unsigned int printEachElement = parameterSet_.get<unsigned int>("print_each_element");
	    unsigned int counter =1;
	    for (const auto& e : elements(basis_.gridView()))
	    {
	    	if (counter % printEachElement == 1 )
	        	std::cout << "assemble stiffness matrix - element: " << counter << std::endl;
	        localView.bind(e);
	    	muLocal.bind(e);
    auto B_Term = prestrainImp
	    	lambdaLocal.bind(e);

	    	Matrix<FieldMatrix<double,1,1> >  elementMatrix;                                          // type ?
	    	computeLocalStiffnessMatrix(localView, muLocal, lambdaLocal, elementMatrix);

	    	for (size_t i=0; i<elementMatrix.N(); i++){
	       		auto row = localView.index(i);
	      		for (size_t j=0; j<elementMatrix.M(); j++ ){
	        		auto col = localView.index(j);
	        		matrix[perIndexMap_[row]][perIndexMap_[col]] += elementMatrix[i][j];
	      		}
	    	}
	    	counter += 1;
	    }
	}


	void assembleLoadVector(const Func2Tensor& matrixFieldFunc, VectorCT& rhs)
	{
		std::cout << "assemble load vector.\n";

	  	//rhs.resize(basis_.size());
		rhs.resize(sizeBasisTorus_);

	  	auto localView = basis_.localView();
	  	auto loadGVF  = makeGridViewFunction(matrixFieldFunc, basis_.gridView());
	  	auto loadFunctional = localFunction(loadGVF);

	  	auto muGVF  = makeGridViewFunction(mu_, basis_.gridView());
	    auto muLocal = localFunction(muGVF);
	    auto lambdaGVF  = makeGridViewFunction(lambda_, basis_.gridView());
	    auto lambdaLocal = localFunction(lambdaGVF);

	  	unsigned int printEachElement = parameterSet_.get<unsigned int>("print_each_element");
	  	unsigned int counter =1;
	  	for (const auto& e : elements(basis_.gridView()))
	  	{
	    	if (counter % printEachElement == 1)
	      		std::cout << "assemble load vector: next element. " << counter << std::endl;
	    	// Bind the local FE basis view to the current element
	    	localView.bind(e);
	    	loadFunctional.bind(e);
	    	muLocal.bind(e);
	    	lambdaLocal.bind(e);

	    	VectorCT elementRhs;
	    	computeLocalLoad(localView, muLocal, lambdaLocal, loadFunctional, elementRhs);

	    	for (size_t i=0; i<elementRhs.size(); i++) {
	      	  auto row = localView.index(i);
	      	  rhs[perIndexMap_[row]] += elementRhs[i];
	      	  /*log_ << "\ni: " << i
	      	  	<< "\nrow: " << row
	      	  	<< "\nperIndexMap_[row]: " << perIndexMap_[row]
	      	  	<< "\nelementRhs[i]: " << elementRhs[i] << std::endl;*/
	    	}
	    	counter += 1;
	  }
	}


	ScalarRT energy(const Func2Tensor& matrixFieldFunc1,
                                   const Func2Tensor& matrixFieldFunc2)
	{
		ScalarRT energy = 0.0;

	  	auto localView = basis_.localView();

	  	auto matrixField1GVF  = makeGridViewFunction(matrixFieldFunc1, basis_.gridView());
	  	auto matrixField1 = localFunction(matrixField1GVF);
	  	auto matrixField2GVF  = makeGridViewFunction(matrixFieldFunc2, basis_.gridView());
	  	auto matrixField2 = localFunction(matrixField2GVF);

	  	auto muGridF  = makeGridViewFunction(mu_, basis_.gridView());
	    auto muLocal = localFunction(muGridF);
	    auto lambdaGridF  = makeGridViewFunction(lambda_, basis_.gridView());
	    auto lambdaLocal = localFunction(lambdaGridF);

	  	for (const auto& e : elements(basis_.gridView()))
	  	{
		    localView.bind(e);
		    matrixField1.bind(e);
		    matrixField2.bind(e);
		    muLocal.bind(e);
		    lambdaLocal.bind(e);

		    FieldVector<double,1> elementEnergy(0);

		    auto geometry = e.geometry();
		    const auto& localFiniteElement = localView.tree().child(0).finiteElement();

		    int order = 2*(dim*localFiniteElement.localBasis().order()-1);
		    const QuadratureRule<double, dim>& quad = QuadratureRules<double, dim>::rule(e.type(), order);

		    for (size_t pt=0; pt < quad.size(); pt++) {
		      const Domain& quadPos = quad[pt].position();
		      const double integrationElement = geometry.integrationElement(quadPos);

		      auto strain1 = matrixField1(quadPos);

		      // St.Venant-Kirchhoff stress
		      MatrixRT stressU(0);
		      stressU.axpy(2*muLocal(quadPos), strain1); //this += 2mu *strainU        // eigentlich this += 2mu *strain1 ?

		      double trace = 0;
		      for (int ii=0; ii<nCompo; ii++)
		        trace += strain1[ii][ii];

		      for (int ii=0; ii<nCompo; ii++)
		        stressU[ii][ii] += lambdaLocal(quadPos) * trace;

		      auto strain2 = matrixField2(quadPos);

	  	  	  //log_ << "strain2= \n";
	  		  //for (int iii = 0; iii < 3; iii++)
	  	  	  	//log_ << strain2[iii] << std::endl;

		      // Local energy density: stress times strain
		      double energyDensity = 0;
		      for (int ii=0; ii<nCompo; ii++)
		        for (int jj=0; jj<nCompo; jj++)
		          energyDensity += stressU[ii][jj] * strain2[ii][jj];

		      //elementConstantTerm += linElasTensor4(localMatrixField1(quadPos), localMatrixField2(quadPos), mu(quadPos), lambda(quadPos)) * quad[pt].weight() * integrationElement;
		        elementEnergy += energyDensity * quad[pt].weight() * integrationElement;
		    }
		    energy += elementEnergy;
		  }
		  return energy;
	}


	int assemblePeriodicIndexMap()     // see Sander-DuneBook p.313..determine positions of Lagrange-nodes
	{
	  std::vector<double> verticesL;
	  auto fL = [](const Domain& x){return FieldVector<double,1>{x[0]};};
	  Functions::interpolate(basis_, verticesL, fL);

	  std::vector<double> verticesX2;
	  auto fAcrossFromX2 = [](const Domain& x){return FieldVector<double,1>{x[1]};};
	  Functions::interpolate(basis_, verticesX2, fAcrossFromX2);

	  std::vector<double> verticesX3;
	  auto fAcrossFromX3 = [](const Domain& x){return FieldVector<double,1>{x[2]};};
	  Functions::interpolate(basis_, verticesX3, fAcrossFromX3);

	  int sizeBasisTorus = 0;
	  for (int d=0; d < nCompo; d++)
	    for (int i = d*basis_.size()/nCompo; i < (d+1)*verticesL.size()/nCompo; i++)
	      if (verticesL[i]!=1.0){
	        perIndexMap_[i]=sizeBasisTorus;
	        sizeBasisTorus++;
	      }else{
	        auto iL =  oppositeVertice(d, verticesL, verticesX2, verticesX3, i);
	        perIndexMap_[i]=perIndexMap_[iL];
	      }
	  log_ << "sizeBasisTorus: " << sizeBasisTorus << std::endl;
	  std::cout << "sizeBasisTorus: " << sizeBasisTorus << std::endl;
	  if (parameterSet_.get<bool>("plot_periodic_index_mapper")) {
	  	log_ << "\nplot_periodic_index_mapper:\n ";
	  	for (int i = 0; i < perIndexMap_.size(); i++){
	  		log_ << i << " " << perIndexMap_[i] << std::endl;
	  	}
	  }

	  return sizeBasisTorus;
	}


	int oppositeVertice(const int d,
                    	const std::vector<double>& verticesL,
                    	const std::vector<double>& verticesAcrossFromX2 ,
                    	const std::vector<double>& verticesAcrossFromX3,
                    	int i)
	{
	  int j=0;
	  for (int iL = d*verticesL.size()/nCompo; iL < (d+1)*verticesL.size()/nCompo; iL++)
	    if (verticesL[iL]==0.0 && verticesAcrossFromX2[i]==verticesAcrossFromX2[iL] && verticesAcrossFromX3[i]==verticesAcrossFromX3[iL])
	      j = iL;
	  return j;
  // test nodeL[i]==1.0
  // test d < nCompo
	}


	void getOccupationPattern(MatrixIndexSet& occupationPattern)
	{
		// in flat interleaved ist occupationPattern ein großer block statt 9, . erste 3x3 block in occupationPattern wuerde lauten:
		  // 1x1x   1x1y    1x1z
		  // 1y1x   1y1y    1y1z
		  // 1z1x   1z1y    1z1z
		  // Machen aber lexicographic
		  std::cout << "setting non zero entries for stiffness matrix begins.\n";

		  occupationPattern.resize(sizeBasisTorus_,sizeBasisTorus_);

		  auto localView = basis_.localView();
		  for(const auto& e : elements(basis_.gridView()))
		  {
		    localView.bind(e);
		    for (size_t i =0; i<localView.size(); i++) {
		      auto iIdx = localView.index(i);
		      for (size_t j=0; j<localView.size(); j++) {
		        auto jIdx = localView.index(j);
		        occupationPattern.add(perIndexMap_[iIdx],perIndexMap_[jIdx]);
		      }
		    }
		  }

		  auto nE = basis_.size();
		  auto nNodeBasis = nE/nCompo;

		  if(parameterSet_.get<bool>("set_one_base_function_to_0" )){                                                       // warum wird dieser Teil benötigt?
		  	for (int d=0; d < nCompo; d++)
				for (int j = 0; j < nNodeBasis; j++)
			        occupationPattern.add(perIndexMap_[0 + d*nNodeBasis], perIndexMap_[0 + d*nNodeBasis]);                  // j ????!!
		  }

		  if(parameterSet_.get<bool>("set_integral_0")){
		  	for (int d=0; d < nCompo; d++)
				for (int j = 0; j < nNodeBasis; j++)
			        occupationPattern.add(perIndexMap_[0 + d*nNodeBasis], perIndexMap_[j + d*nNodeBasis]);
		  }

		  if(parameterSet_.get<bool>("set_first_moment_0"))
		   	for (int j = 0; j < nE; j++)
		       	occupationPattern.add(perIndexMap_[nNodeBasis] + 1, perIndexMap_[j]);

		  std::cout << "finished occupation pattern\n";
	}


	void setOneBaseFunctionToZero (MatrixCT& stiffnessMatrix,
                           VectorCT& load_a,
                           VectorCT& load_K1, VectorCT& load_K2, VectorCT& load_K3)
	{
		if (parameterSet_.get<bool>("set_one_base_function_to_0")){
			log_ << "\nassembling: one base function to zero\n";
			std::cout << "\nassembling: one base function to zero\n";
			auto nE = basis_.size();
			auto nENode = nE/nCompo;

			//arbitrary index < nENode
			int arbitraryIndex = 0;
			for (int d = 0; d < dim; d++){
				auto row = perIndexMap_[arbitraryIndex + d * nENode];                  // in jeder Komponente auf 0 gesetzt
				load_a[row] = 0.0;
		    	load_K1[row] = 0.0;
		    	load_K2[row] = 0.0;
		    	load_K3[row] = 0.0;
		    	auto cIt    = stiffnessMatrix[row].begin();
		    	auto cEndIt = stiffnessMatrix[row].end();
		    	for (; cIt!=cEndIt; ++cIt)
		      		*cIt = (cIt.index()==row) ? 1.0 : 0.0;
			}
		}


	}



	void setIntegralZero (MatrixCT& stiffnessMatrix,
                           VectorCT& load_a,
                           VectorCT& load_K1, VectorCT& load_K2, VectorCT& load_K3)
	{

	    //
	    //  0. moment invariance, 3 dofs
	    //
		if (parameterSet_.get<bool>("set_integral_0")){
			log_ << "\nassembling: integral to zero\n";
			std::cout << "\nassembling: integral to zero\n";
		    auto n = basis_.size();
		    auto nNodeBasis = n/nCompo;
		    auto localView = basis_.localView();

			int arbitraryIndex = 0;
	        for (int d = 0; d < dim; d++) {
	            auto row = perIndexMap_[arbitraryIndex + d*nNodeBasis];
	            stiffnessMatrix[row] = 0;
	            load_a[row] = 0.0;
			    load_K1[row] = 0.0;
			    load_K2[row] = 0.0;
			    load_K3[row] = 0.0;
	        }

	        unsigned int elementCounter = 1;
	        for (const auto& e : elements(basis_.gridView()))
	        {
	          if (elementCounter % 50 == 1)                                                                           // ?????
	            std::cout << "integral = zero - treatment - element: " << elementCounter << std::endl;
	          localView.bind(e);

	          BlockVector<FieldVector<double,1> > elementColumns;
	          elementColumns.resize(localView.size());
	          elementColumns = 0;

	          const auto& localFiniteElement = localView.tree().child(0).finiteElement();
	          const auto nSf = localFiniteElement.localBasis().size();

	          int order = 2*(dim*localFiniteElement.localBasis().order()-1);
	          const QuadratureRule<double, dim>& quad = QuadratureRules<double, dim>::rule(e.type(), order);

	          for (size_t pt=0; pt < quad.size(); pt++) {

	            const FieldVector<double,dim>& quadPos = quad[pt].position();
	            const double integrationElement = e.geometry().integrationElement(quadPos);

	            std::vector<FieldVector<double,1> > shapeFunctionValues;
	            localFiniteElement.localBasis().evaluateFunction(quadPos, shapeFunctionValues);

	            for (size_t k=0; k<nSf; k++)
	              for (size_t d=0; d<nCompo; d++)
	                elementColumns[ k + d*nSf] += shapeFunctionValues[k] * quad[pt].weight() * integrationElement;
	          }

	          for (int d = 0; d < dim; d++) {
	            auto row = perIndexMap_[arbitraryIndex + d*nNodeBasis];
	            for (size_t j=0; j<nSf; j++){
	              auto col = perIndexMap_[ (localView.index(j+ d*nSf)) ];
	              stiffnessMatrix[row][col] += elementColumns[j+d*nSf];
	            }
	          }
	          elementCounter ++;
	        }//end for each element
		}

	}


	void setFirstMomentZero (MatrixCT& stiffnessMatrix,
                           VectorCT& load_a,
                           VectorCT& load_K1, VectorCT& load_K2, VectorCT& load_K3)
	{
	  if (parameterSet_.get<bool>("set_first_moment_0")){
	  	  log_<< "\nassembling: first moment = zero\n";
	  	  std::cout<< "\nassembling: first moment = zero\n";

		  auto zerox2x3Term = [] (const Domain& z) { return Domain({0, z[1], z[2]}); }; //Range components = Domain compo Y X2 X3                           // ???
		  auto zerox2x3  = Functions::makeGridViewFunction(zerox2x3Term, basis_.gridView());

		  auto n = basis_.size();
		  auto nNodeBasis = n/nCompo;
		  auto localView = basis_.localView();
		  auto zerox2x3Local = localFunction(zerox2x3);

		  //arbitrary index < nENode
		  int arbitraryIndex = nNodeBasis; //setOneBaseFunctionTo0 takes index= 0, we hav to modify rows which affect 2. or 3. component
	      auto row = perIndexMap_[arbitraryIndex] + 1;                                                                                                      // ???
	      stiffnessMatrix[row] = 0;
	      load_a[row] = 0;
	      load_K1[row] = 0;
	      load_K2[row] = 0;
	      load_K3[row] = 0;

	      unsigned int elementCounter = 1;
	      for (const auto& e : elements(basis_.gridView()))
	      {
	        if (elementCounter % 50 == 1)                                                                                                                  // ????
	          std::cout << "first moment = zero - element: " << elementCounter << std::endl;

	        localView.bind(e);
	        zerox2x3Local.bind(e);

	        BlockVector<FieldVector<double,1> > elementColumns;
	        elementColumns.resize(localView.size());
	        elementColumns = 0;
	        const auto& localFiniteElement = localView.tree().child(0).finiteElement();
	        const auto nSf = localFiniteElement.localBasis().size();
	        int order = 2*(dim*localFiniteElement.localBasis().order()-1);
	        const QuadratureRule<double, dim>& quad = QuadratureRules<double, dim>::rule(e.type(), order);

	        for (size_t pt=0; pt < quad.size(); pt++) {

	          const FieldVector<double,dim>& quadPos = quad[pt].position();
	          const double integrationElement = e.geometry().integrationElement(quadPos);
	          std::vector<FieldVector<double,1> > shapeFunctionValues;
	          localFiniteElement.localBasis().evaluateFunction(quadPos, shapeFunctionValues);

	          for (size_t k=0; k<nSf; k++)
	            for (size_t d=0; d<dim; d++){
	              FieldVector<double,3> sFVVector(0);
	              sFVVector[d] = shapeFunctionValues[k];
	              size_t ll = localView.tree().child(d).localIndex(k);
	              elementColumns[ll] += (sFVVector * zerox2x3Local(quadPos)) * quad[pt].weight() * integrationElement;
	            }
	        }

	        for (int j = 0; j < localView.size(); j++){
	          auto col = perIndexMap_[(localView.index(j))];
	          stiffnessMatrix[row][col] += elementColumns[j]; //arg1 = 1...secound vertice, arg2 = 0...first dimension
	        }
	        elementCounter ++;
	      }//end for each element

	  } // end if
	}





	// main method which uses the previous ones

	void assemble(MatrixCT& stiffnessMatrix,
	VectorCT& load_a, VectorCT& load_K1,
	VectorCT& load_K2, VectorCT& load_K3)
	{

	    MatrixIndexSet occupationPattern;
	    getOccupationPattern(occupationPattern);
	    occupationPattern.exportIdx(stiffnessMatrix);

	  	assembleStiffnessMatrix(stiffnessMatrix);

	  	// Forces for cell problems = strain basis functions
	    // Basis for the limiting Biot strain:
	    // E_h = 1/h (sqrt(F^T F) - Id) -> E(U,K) + G
	    // E(U,K) = sym( K * (0,x2,x3)^T + U  ot e1)

	    // lateral stretch strain	E_U1 = e1 ot e1
	  	Func2Tensor E_a = [] (const Domain& z) {
	  		 return MatrixRT{{1.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}}; };
	      						//{{0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 1.0}};

	  	//  x2   0  0   |    x3   0  0  |     0        1/2 x3   -1/2 x2
	  	//  0    0  0   |    0    0  0  |     1/2 x3   0             0
	  	//  0    0  0   |    0    0  0  |    -1/2 x2   0             0
	  	//
	  	//  Ki.. basis of Skew(3) ... rod curvature
	  	//
	    // twist in x2-x3 plane E_K1 = sym (e1 x (0,x2,x3)^T ot e1)
	    Func2Tensor E_K1 = [] (const Domain& z) {
	      return MatrixRT{{0.0  , -0.5*z[2],      0.5 * z[1]},
	                          {-0.5*z[2], 0.0  ,       0.0 },
	                          {0.5*z[1], 0.0  ,       0.0}}; };
	                          //{{0.0, 0.5*z[1],-0.5*z[0]}, {0.5*z[1],0.0, 0.0}, {-0.5*z[0], 0.0, 0.0}};

	  	//  bend strain in x1-x2 plane	E_K2 = sym (e2 x (0,x2,x3)^T ot e1)
	  	Func2Tensor E_K2 = [] (const Domain& z) {
	      return MatrixRT{{z[2], 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0}}; };
	      					//{{z[0], 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};

	  	// bend strain in x1-x3 plane	E_K3 = sym (e3 x (0,x2,x3)^T ot e1)
	  	Func2Tensor E_K3 = [] (const Domain& z) {
	      return MatrixRT{{-z[1], 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}}; };
	      						//{{z[1], 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};

	    assembleLoadVector(E_a, load_a);
	    //log_ << "\n\n\n\n\n\n";
	    assembleLoadVector(E_K1, load_K1);
	    assembleLoadVector(E_K2, load_K2);
	    assembleLoadVector(E_K3, load_K3);

	  	// invariants treatments
	    setOneBaseFunctionToZero(stiffnessMatrix, load_a, load_K1, load_K2, load_K3);
	    setFirstMomentZero(stiffnessMatrix, load_a, load_K1, load_K2, load_K3);
	    setIntegralZero(stiffnessMatrix, load_a, load_K1, load_K2, load_K3);
	};

}; // end class


#endif



// shear strain	in x2  E_U2 = sym (e2 ot e1)
	    /*
	  	F E_U2Term = [] (const Domain& z) {
	  		MatrixRT ret = {{0.0, 0.5, 0.0}, {0.5, 0.0, 0.0}, {0.0, 0.0, 0.0}};
	  		return ret; };

	  	// shear strain	in x2  E_U2 = sym (e3 ot e1)
	  	F E_U3Term = [] (const Domain& z) {
	  		MatrixRT ret = {{0.0, 0.0, 0.5}, {0.0, 0.0, 0.0}, {0.5, 0.0, 0.0}};
	  		return ret; };
	    */
